from lark import Tree
from os import listdir
from os.path import join
from swissql.analyzers.rule_checker.custom_rule import CustomRule
from swissql.analyzers.rule_checker.tree_pattern_find import TreePatternFinder
from swissql.utils.exceptions import Error


class RuleFinder:
    """
    Class that is used to locate user defined patterns

    ...

    Attributes
    ----------
    SPLITTER: str
        Constant string that is used to split comment from grammar in rule file
    PATTERN: str
        Name of pattern to be searched in query
    rules: list[CustomRule]
        List of user defined rules

    Methods
    -------
    __parse_rule_file(cls, data: str)
        Function that parses rule file data
    load_rule(cls, rule_file: str)
        Loads specified rule
    load_rules(cls, rule_files: list[str])
        Loads list of user specified rules
    find_rules(cls, sql: str)
        Finds patterns in query and 
    """
    SPLITTER: str = ':END_OF_COMMENT:'
    PATTERN: str = 'PATTERN'

    rules: list[CustomRule] = []

    @classmethod
    def __parse_rule_file(cls, data: str) -> tuple[str]:
        try:
            comment, grammar = data.split(cls.SPLITTER)
            return (comment, grammar)
        except ValueError as e:
            raise Error("invalid rule file format") from e

    @classmethod
    def load_rule(cls, rule_file: str):
        data = ""
        with open(rule_file, mode="r") as f:
            data = f.read()
        comment, grammar = cls.__parse_rule_file(data)
        cls.rules.append(CustomRule(name=rule_file, grammar=grammar, comment=comment))

    @classmethod
    def load_rules(cls, rule_files: list[str]):
        for rule_file in rule_files:
            cls.load_rule(rule_file)

    @classmethod
    def find_rules(cls, sql: str) -> list[(CustomRule, list[(int, int)])]:
        found = list()
        for rule in cls.rules:
            try:
                tree = rule.parse_sql_tree(sql)
            except Exception as e:
                continue
            finder = TreePatternFinder()
            finder.visit(tree)
            if len(finder.positions) > 0:
                found.append((rule, finder.positions))
        return found



from sqlglot import parse_one
from sqlglot.errors import ParseError
from swissql.utils.exceptions import Error


class SqlParser:
    # parser: Lark

    @classmethod
    def parse_tree(cls, sql: str) -> str:
        try:
            tree = parse_one(sql, read="spark")
        except ParseError as e:
            raise Error("Spark SQL parse error") from e
        return repr(tree)

    # @classmethod
    # def initialize(cls, lexer='basic'):
    #     grammar = ''
    #     with open('grammar/sql.lark', mode='r') as f:
    #         grammar = f.read()
    #     parser = Lark(grammar, start='start', lexer=lexer)

    # @classmethod
    # def parse_spark_sql(cls, sql: str) -> Tree:
    #     tree = cls.parser.parse(sql)
    #     return tree

    # @classmethod
    # def parse_spark_sql_as_json(cls, sql: str):
    #     tree = cls.parse_spark_sql(sql)
    #     output = TreeToJson().transform(tree)
    #     return output

    # @classmethod
    # def parse_spark_sql_as_str(cls, sql: str) -> str:
    #     tree = cls.parse_spark_sql(sql)
    #     output = tree.pretty()
    #     return output



import json
import sys
import os
from argparse import ArgumentParser, Namespace, BooleanOptionalAction
from swissql.analyzers.style.sql_style_error import SqlfluffCheck
from swissql.analyzers.anti_pattern.anti_pattern_finder import AntiPatternFinder
from swissql.analyzers.syntax.sql_parser import SqlParser
from swissql.analyzers.style.sql_formatter import SqlFormatter
from swissql.utils.schema_reader import SchemaReader
from swissql.utils.exceptions import Error, exception_handler
from swissql.analyzers.optimizations.static_optimizer import StaticOptimizer
from swissql.analyzers.optimizations.dynamic_optimizer import DynamicOptimizer
from swissql.analyzers.rule_checker.rule_finder import RuleFinder
from swissql.analyzers.rule_checker.custom_rule import CustomRule
from swissql.analyzers.rule_checker.sql_finder import SqlFinder
from swissql.manifest import Manifest
from swissql.analyzers.optimizations.dynamic_optimizer import TableSize

class ArgParser:
    parser: ArgumentParser(prog="swissql")
    args: Namespace
    modes: list[str] = ["syntax", "format", "optimize", "style", "anti_pattern", "rule"]

    @staticmethod
    def __pair_or(argument: str) -> bool:
        return argument not in sys.argv

    @staticmethod
    def __pair_and(argument: str) -> bool:
        return argument in sys.argv

    @classmethod
    def __get_queries(cls) -> list[str]:
        queries = None
        if cls.args.q:
            queries = cls.args.q.split(";")
        elif cls.args.x:
            SqlFinder.initialize()
            if not cls.args.f:
                raise Error("no file provided")
            filename = cls.args.f
            print("\u001b[33m[Extracting Spark SQLs from file using lark]\u001b[0m")
            sqls = SqlFinder.extract_sql_from_file(filename)
            if sqls:
                print("Found:")
                for sql in sqls:
                    print(sql)
            else:
                print("Did not find any SQLs")
            queries = sqls
        elif cls.args.f:
            query = ""
            try:
                with open(cls.args.f, mode="r") as f:
                    query = f.read()
            except FileNotFoundError as e:
                raise Error("query file not found") from e
            queries = query.split(";")
        else:
            raise Error("either -q or -f argument is required")
        queries = list(map(lambda x: x.strip(), queries))
        queries = list(filter(lambda x: x, queries))
        return queries

    @classmethod
    def initialize(cls) -> None:

        static_optimizers: list[str] = StaticOptimizer.get_optimizers()
        dynamic_optimizers: list[str] = DynamicOptimizer.get_optimizers()
        cls.parser = ArgumentParser(
            prog=Manifest.APP_NAME,
            description=Manifest.APP_DESCRIPTION,
        )
        cls.parser.add_argument(
            "mode", choices=cls.modes + ["all", "construct"], help="mode of operation"
        )
        cls.parser.add_argument(
            "-q",
            # "--query",
            required=cls.__pair_or("-f"),
            help="specify SQL query",
        )
        cls.parser.add_argument(
            "-f",
            # "--file-sql",
            required=cls.__pair_or("-q"),
            help="specify file to read SQL query from",
        )
        cls.parser.add_argument(
            "-x",
            # "--extract",
            action=BooleanOptionalAction,
            help="specify if queries should be extracted from source files",
        )
        cls.parser.add_argument(
            "-s",
            # "--schema",
            help="""specify schema:
        Schema is a mapping in one of
            the following forms:\n
                1. {table: {col: type}}\n
                2. {db: {table: {col: type}}}\n
                3. {catalog: {db: {table: {col: type}}}}\n
        Example types: INT, STRING""",
        )
        cls.parser.add_argument(
            "-F",
            # "--File-schema",
            help="specify schema file to read schema from",
        )
        cls.parser.add_argument(
            "-o",
            #  "--optimizer",
            choices=static_optimizers + dynamic_optimizers,
            help="specify optimizers",
        )
        cls.parser.add_argument(
            "-r",
            # "--rule",
            help='specify rule files as list',
            action='append'
        )
        cls.parser.add_argument(
            "-c",
            # "--choice",
            choices=cls.modes,
            help="specify modes for analysis constructor",
            action="append",
        )
        cls.parser.add_argument(
            "-t",
            # "--table"
            help="specify table sized as json file",
        )
        cls.parser.add_argument(
            "--rules-sqlfluff",
            help="specify rules file for sqlfluff",
        )
        # cls.parser.add_argument("-c", help="specify")
        cls.parser.add_argument(
            "--dialect", default="sparksql", help="specify sqlfluff dialect"
        )
        cls.parser.add_argument(
            "--output-mode",
            default="str",
            choices=["str", "json"],
            help="specify output format",
        )
        cls.args = cls.parser.parse_args()

    @classmethod
    @exception_handler()
    def analyze_one(cls, query: str, mode: str) -> None:
        if mode == "syntax":
            try:
                print("\u001b[33m[Generating syntax tree using sqlglot]\u001b[0m")
                output = SqlParser.parse_tree(query)
                print(output)
            except Exception as e:
                 print(f"\u001b[31m{Manifest.APP_NAME}: error: {e}\u001b[0m")
        elif mode == "format":
            try:
                print("\u001b[33m[Formatting sql query using sqlglot]\u001b[0m")
                output = SqlFormatter.format_one(query)
                print(output)
            except Exception as e:
                 print(f"\u001b[31m{Manifest.APP_NAME}: error: {e}\u001b[0m")
        elif mode == "style":
            try:
                print("\u001b[33m[Style sql query use sqlfluff]\u001b[0m")
                # получаем ссылку на файл с помощью lark
                # print(cls.args.f)
                output = SqlfluffCheck(
                    "lint",
                    file=cls.args.f,
                    q=cls.args.q,
                    rules=cls.args.rules_sqlfluff,
                    dialect=cls.args.dialect,
                ).create_str()
                print(output)
                #
                # Меняем позиции в тексте
            except Exception as e:
                 print(f"\u001b[31m{Manifest.APP_NAME}: error: {e}\u001b[0m")
        elif mode == "optimize":
            try:
                optimization = cls.args.o
                schema = None

                if optimization in ["optimize", "qualify_columns"]:
                    if cls.args.s:
                        schema = SchemaReader.parse_json_schema(cls.args.s)
                    elif cls.args.F:
                        schema = SchemaReader.parse_file_json_schema(cls.args.F)
                    else:
                        raise Error("schema is not provided")
                if optimization in StaticOptimizer.get_optimizers():
                    print(
                        "\u001b[33m[Optimizing staticaly sql query using sqlglot]\u001b[0m"
                    )
                    print(f"Optimization: {cls.args.o}")
                    optimized = StaticOptimizer.optimize(cls.args.o, query, schema)
                    print(optimized)
                    print()
                if optimization in DynamicOptimizer.get_optimizers():
                    table_sizes = cls.args.t
                    if not table_sizes:
                        raise Error("no table sizes specified")
                    try:
                        data = ""
                        with open(cls.args.t, mode="r") as f:
                            data = f.read()
                        table_sizes = json.loads(data)
                        table_sizes = [TableSize(**x) for x in table_sizes]
                    except FileNotFoundError as e:
                        raise Error("file with table sizes not found")
                    except Exception as e:
                        raise Error("dict parse error")
                    print("\u001b[33m[Optimizing dynamicaly sql query using lark]\u001b[0m")
                    print(f"Optimization: {cls.args.o}")
                    # try:
                    hints = DynamicOptimizer.optimize(cls.args.o, query, table_sizes)
                    # except Exception as e:
                    #     raise Error('invalid table_size format') from e
                    for hint in hints:
                        print(f"Hint: {hint[0]}")
                        print(f"Position: ({hint[1][0]}, {hint[1][1]})")
            except Exception as e:
                print(f"\u001b[31m{Manifest.APP_NAME}: error: {e}\u001b[0m")
        elif mode == "anti_pattern":
            try:
                print("\u001b[33m[Detecting anti-patterns using sqlcheck]\u001b[0m")

                # TODO: add verbosity level option
                instance = AntiPatternFinder(verbose=True)
                sqlcheck_output = ""
                if cls.args.f:
                    sqlcheck_output = instance.find_anti_patterns_from_file(cls.args.f)
                elif cls.args.q:
                    sqlcheck_output = instance.find_anti_patterns_from_query(cls.args.q)
                else:
                    raise Error("either -q or -f argument is required")
                print(sqlcheck_output)
            except Exception as e:
                 print(f"\u001b[31m{Manifest.APP_NAME}: error: {e}\u001b[0m")
        elif mode == "rule":
            try:
                print("\u001b[33m[Finding rules using lark]\u001b[0m")
                rules = cls.args.r
                if not rules:
                    raise Error("no rule specified")
                try:
                    RuleFinder.load_rules(rules)
                except Exception as e:
                    print(e)
                    raise Error("rule file is not found")
                found = RuleFinder.find_rules(query)
                for composition in found:
                    rule, positions = composition
                    positions = " ".join([str(el) for el in positions])
                    print(f"Rule {rule.name} found:")
                    print(f"Positions: {positions}")
                    print(f"Comment: {rule.comment}")
                    print()
            except Exception as e:
                 print(f"\u001b[31m{Manifest.APP_NAME}: error: {e}\u001b[0m")
        elif mode == "construct":
            for mode in cls.args.c:
                cls.analyze_one(query, mode)
                print()
        elif mode == "all":
            for mode in cls.modes:
                cls.analyze_one(query, mode)
                print()

    @classmethod
    def analyze_queries(cls) -> None:
        queries = cls.__get_queries()
        for query in queries:
            print(f"\u001b[33m[Analyzing query:\u001b[35m {query}\u001b[33m]\u001b[37m")
            cls.analyze_one(query, cls.args.mode)




from lark import Tree, Visitor


class TreePatternFinder(Visitor):
    positions: list[(int, int)] = []

    def pattern(self, tree):
        self.positions.append((tree.meta.line, tree.meta.column))


class TreeStringFinder(Visitor):
    strings: list[str] = []

    def string(self, tree):
        for el in tree.children:
            self.strings.append(el.value)


class TreeJoinFinder(Visitor):
    join_spots: list[(str, (int, int))] = []

    def from_item_pattern(self, tree):
        try:
            name = tree.children[0].children[0].children[0].value
            meta = tree.children[0].children[0].meta
            position = (meta.line, meta.column)
            self.join_spots.append((name, position))
        except Exception:
            return



from enum import Enum
from typing import Union, Callable, Optional
from sqlglot import Expression, parse_one
from sqlglot.optimizer import optimize
from sqlglot.optimizer.normalize import normalize
from sqlglot.optimizer.eliminate_subqueries import eliminate_subqueries
from sqlglot.optimizer.expand_multi_table_selects import expand_multi_table_selects
from sqlglot.optimizer.isolate_table_selects import isolate_table_selects
from sqlglot.optimizer.optimize_joins import optimize_joins
from sqlglot.optimizer.pushdown_predicates import pushdown_predicates
from sqlglot.optimizer.pushdown_projections import pushdown_projections
from sqlglot.optimizer.qualify_tables import qualify_tables
from sqlglot.optimizer.qualify_columns import qualify_columns
from sqlglot.optimizer.quote_identities import quote_identities
from sqlglot.optimizer.unnest_subqueries import unnest_subqueries


class StaticOptimizer:
    """
    Static optimizer of spark sql query

    ...

    Attributes
    ----------
    optimizers : dict 
        Dict of different optimizers ofr spark sql code

    Methods
    -------
    optimize(cls, optimizer: str, sql: str, schema: Optional[dict] = None)
        Chooses optimizer and using it to generate improved query 
    get_optimizers(cls)
        Return list of available optimizers
    """

    optimizers = {
        "optimize": optimize,
        "normalize": normalize,
        "eliminate_subqueries": eliminate_subqueries,
        "expand_multi_table_selects": expand_multi_table_selects,
        "isolate_table_selects": isolate_table_selects,
        "optimize_joins": optimize_joins,
        "pushdown_predicates": pushdown_predicates,
        "pushdown_projections": pushdown_projections,
        "qualify_tables": qualify_tables,
        "qualify_columns": qualify_columns,
        "quote_identities": quote_identities,
        "unnest_subqueries": unnest_subqueries,
    }

    @classmethod
    def optimize(cls, optimizer: str, sql: str, schema: Optional[dict] = None) -> Expression: 
        """
        Parameters
        ----------
        optimizer: str
            Name of optimizer to be used to optimize spark sql query
        sql: str
            Unoptimized spark sql query
        schema: Optional[dict]
            Dataframe schema to provide additional information before optimization

        Return
        ------
        optimized: Expression
            Optimized spark sql query
        """
        expr = parse_one(sql, read="spark")
        optimizer = cls.optimizers[optimizer]
        # optimized = optimizer(expr,schema).sql(pretty=True)
        if optimizer in [qualify_columns, optimize]:
            print(optimizer(expr, schema))
            optimized = optimizer(expr, schema).sql(pretty=True)
        else:
            optimized = optimizer(expr).sql(pretty=True)
        return optimized

    @classmethod
    def get_optimizers(cls) -> list[str]:
        """
        Parameters
        ----------
        None

        Return
        ------
        keys: list[str]
            List of available optimizers
        """
        return list(cls.optimizers.keys())



from sqlglot import transpile


class SqlFormatter:
    @classmethod
    def format_one(cls, sql: str) -> str:
        statements = transpile(sql,read='spark', write='spark', pretty=True)
        return statements[0]



import os

from swissql.utils.exceptions import Error


class AntiPatternFinder:
    def __init__(self, verbose: bool = True):
        """
        Initialize the AntiPatternFinder class
        AntiPatternFinder is a class that can be used to find anti-patterns in sql statements
        It uses the sqlcheck command to find anti-patterns

        :param verbose: if True, verbose flag will be passed to sqlcheck

        """
        self.verbose: bool = verbose
        self.temp_sql_file: str = "temp.sql"

        # Check sqlcheck availability in system
        self.isavailable = os.system("sqlcheck -version") == 0
        if not self.isavailable:
            raise Error("sqlcheck not found in system")

    def find_anti_patterns_from_file(self, sql_filename: str) -> str:
        """
        Find anti-patterns in sql statement from file

        :param sql_filename: sql file name
        :return: sqlcheck output
        """
        # Check availability of sqlcheck
        if not self.isavailable:
            raise Error("sqlcheck not found in system")

        # Run sqlcheck on sqlfile
        # Timeout in case of wrong sqlfile
        # TODO: switch timeout 1 to subprocess timeout to improve crossplatform compatibility
        sqlcheck_output = os.popen(
            f"timeout 1 sqlcheck -c {'-v' if self.verbose else ''} -f { sql_filename }"
        ).read()
        return sqlcheck_output

    def find_anti_patterns_from_query(self, sql: str) -> str:
        """
        Find anti-patterns in sql statement

        :param sql: sql statement
        :return: sqlcheck output
        """
        # Check availability of sqlcheck
        if not self.isavailable:
            raise Exception("sqlcheck not found in system")

        # Using writing to file to provide crossplatform compatibility
        with open(self.temp_sql_file, "w") as f:
            f.write(sql)
        sqlcheck_output = self.find_anti_patterns_from_file(self.temp_sql_file)
        os.remove(self.temp_sql_file)

        return sqlcheck_output

    @staticmethod
    def sqlcheck_output_to_json(sqlcheck_output: str) -> str:
        """?"""
        pass


if __name__ == "__main__":
    # Use example
    instance = AntiPatternFinder(verbose=True)
    print(instance.find_anti_patterns_from_query("select * from table"))
    print(instance.find_anti_patterns_from_file("../../../examples/example2.sql"))



from typing import List


class Manifest:
    APP_NAME: str = "swissql"
    APP_DESCRIPTION: str = "Spark SQL analytic engine"
    APP_VERSION: str = "v0.2.5.4"
    LICENSE: str = "GNU"
    EMAIL: str = "andrey24072002@bk.ru"
    REQUIREMENTS: List = [
        "lark>=1.1.2",
        "sqlfluff>=1.2.1",
        "sqlglot>=4.2.9",
        "pydantic>=1.9.1",
    ]
    AUTHORS_GIT: str = "husker Nicialy Quakumei"
    GIT_URL: str = "https://github.com/IPROSpark/SparkSQL-Analyzer"



import re
from typing import Union
from lark import Tree, Lark
from os import listdir
from os.path import join
from swissql.analyzers.rule_checker.custom_rule import CustomRule
from swissql.analyzers.rule_checker.tree_pattern_find import TreeStringFinder
from swissql.utils.exceptions import Error


class SqlFinder:
    python_parser: Lark
    sql_parser: Lark

    @classmethod
    def __get_parser_from_filename(cls, filename: str) -> Lark:
        return cls.python_parser

    @classmethod
    def initialize(cls):
        # python_grammar_file = 'swissql/analyzers/rule_checker/grammar/python_peg.lark'
        # python_grammar = ''
        # with open(python_grammar_file, mode='r') as f:
        #     python_grammar = f.read()
        # cls.python_parser = Lark(
        #     python_grammar,
        #     parser='lalr',
        #     lexer='basic',
        #     start='start',
        # )
        sql_grammar_file = "swissql/analyzers/rule_checker/grammar/sql.lark"
        sql_grammar = ""
        with open(sql_grammar_file, mode="r") as f:
            sql_grammar = f.read()
        cls.sql_parser = Lark(
            sql_grammar,
            parser="lalr",
            lexer="basic",
            start="start",
            propagate_positions=True,
        )

    @classmethod
    def extract_sql_from_file(cls, filename: str) -> list[str]:
        data = ""
        try:
            with open(filename, mode="r") as f:
                data = f.read()
        except FileNotFoundError as e:
            raise Error("file not found")

        # Extract strings from data
        strings = re.findall(r"('''|\"\"\"|[\"'])((?:\\.|.)*?)\1", data)
        sqls = list()
        for string in strings:
            string = "".join(string)
            string = string.strip('"').strip("'").strip(";")
            try:
                tree = cls.sql_parser.parse(string)
            except Exception as e:
                continue
            sqls.append(string)
        return sqls



from typing import Optional
from lark import Lark, Tree
from dataclasses import dataclass


@dataclass
class CustomRule:
    """
    Dataclass for user defined rules

    ...

    Attributes
    ----------
    name: str
        Filename of rule file
    grammar: str
        Grammar as pattern for finding user defined rules in query
    comment: str
        User defined comment to acknowledge a problem and/or to give insights to how to fix it
    parser: Optional[Lark]
        Rule parser that is used to find user defined patterns

    Methods
    -------
    __initialise_parser()
        Creates parser based on grammar
    parse_sql_tree(sql: str)
        Parses sql query and returns it's tree
    parse_sql_str(sql: str)
        Parses sql query and returns it's tree as string representation
    """

    name: str
    grammar: str
    comment: str

    parser: Optional[Lark] = None

    def __initialise_parser(self) -> None:
        """
        Parameters
        ----------
        None

        Return
        ------
        None
        """
        self.parser = Lark(
            self.grammar,
            parser='lalr',
            lexer='basic',
            start='start',
            propagate_positions=True, # allows to store node position
        )

    def parse_sql_tree(self, sql: str) -> Tree:
        """
        Attributes
        ----------
        sql: str
            Spark sql query to be parsed

        Return
        ------
        tree: Tree
            Parsed sql query tree
        """
        if not self.parser:
            self.__initialise_parser()
        return self.parser.parse(sql)

    def parse_sql_str(self, sql: str) -> str:
        """
        Attributes
        ----------
        sql: str
            Spark sql query to be parsed

        Return
        ------
        tree: str
            String representation of parsed tree
        """
        if not self.parser:
            self.__initialise_parser()
        return repr(self.parser.parse(sql))



from lark import Lark, Tree
from enum import Enum
from typing import Union, Callable, Optional
from sqlglot import Expression, parse_one
from swissql.analyzers.rule_checker.tree_pattern_find import TreeJoinFinder
from pydantic import BaseModel

class TableSize (BaseModel):
    name : str
    size : float

def dynamic_optimize_join(expr: Tree, table_size: list[TableSize]) -> list[(str, (int, int))]:
    finder = TreeJoinFinder()
    finder.visit(expr)
    spots = finder.join_spots
    hints = []
    
    dic = {}
    for table in table_size:
        dic[table.name] = table.size
    
    for spot in spots:
        name, position = spot
        size = dic[name]
        if size <= 1.5:
            hints.append((f"/*+ BROADCAST({name}) */ ", position))
    return hints


class DynamicOptimizer:

    optimizers = {"dynamic_optimize_joins": dynamic_optimize_join}

    @classmethod
    def optimize(
        cls, optimizer: str, sql: str, table_sizes: list[TableSize]
    ) -> list[(str, (int, int))]:
        sql_grammar_file = "swissql/analyzers/rule_checker/grammar/sql_dynamic.lark"
        sql_grammar = ""
        with open(sql_grammar_file, mode="r") as f:
            sql_grammar = f.read()
        sql_parser = Lark(
            sql_grammar,
            parser="lalr",
            lexer="basic",
            start="start",
            propagate_positions=True,
        )
        expr = sql_parser.parse(sql)
        hints = []
        if optimizer == "optimizer":
            for optimizer in cls.get_optimizers():
                cls.optimize_one(optimizer, expr, table_sizes, hints)
        else:
            cls.optimize_one(optimizer, expr, table_sizes, hints)
        return hints

    @classmethod
    def optimize_one(cls, optimizer: str, expr: Tree, table_sizes: list[TableSize], hints: list):
        optimizer = cls.optimizers[optimizer]
        hint = optimizer(expr, table_sizes)
        if hint:
            hints.extend(hint)

    @classmethod
    def get_optimizers(cls) -> list[str]:
        return list(cls.optimizers.keys())



from functools import wraps
from swissql.manifest import Manifest


class Error(Exception):
    """
    Error class is used to raise exceptions

    """

    def __init__(self, message: str):
        super().__init__()
        self.message = message


def exception_handler():
    def inner_func(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Error as e:
                print(f"\u001b[31m{Manifest.APP_NAME}: error: {e.message}\u001b[0m")

        return wrapper

    return inner_func



from setuptools import setup, find_packages

from swissql.manifest import Manifest
 
 

 
setup(name=Manifest.APP_NAME,
      version=Manifest.APP_VERSION,
      url=Manifest.GIT_URL,
      license=Manifest.LICENSE,
      author=Manifest.AUTHORS_GIT,
      author_email=Manifest.EMAIL,
      packages= find_packages(),
      install_requires = Manifest.REQUIREMENTS ,
      description=Manifest.APP_DESCRIPTION,
      
      python_requires=">=3.7.*",
      entry_points={
        'console_scripts': [
            'swissql = swissql.main:main_start',
        ]}
      )


from json import loads
from json.decoder import JSONDecodeError
from swissql.utils.exceptions import Error


class SchemaReader:
    @staticmethod
    def parse_json_schema(json_schema: str) -> dict:
        try:
            schema = loads(json_schema)
        except JSONDecodeError as e:
            raise Error("Schema is not in json format")

        return schema

    @classmethod
    def parse_file_json_schema(cls, filename: str) -> dict:
        json_schema = ""
        try:
            with open(filename, mode="r") as f:
                json_schema = f.read()
        except FileNotFoundError as e:
            raise Error("Schema file is not found")
        schema = cls.parse_json_schema(json_schema)
        return schema



import os
import pathlib
from typing import Any, Union
from sqlfluff.cli.commands import cli


class SqlfluffCheck:
    def __init__(
        self,
        mode: Union[str, None],
        file: Union[str, None] = None,
        q: Union[str, None] = None,
        rules: Union[str, None] = None,
        dialect: str = None,
        help: Union[str, None] = None,
    ) -> None:
        self.file = file
        self.querry = q
        self.rules = rules
        self.dialect = dialect
        self.mode = mode
        self.help = help
        # TODO пофиксить создание файла
        self.filename = (
            pathlib.Path(__file__)
            .parent.resolve()
            .joinpath(pathlib.Path("intermediate.sql"))
        )

    def create_str(self) -> str:
        string = "sqlfluff"
        if self.help:
            return string + " -" + self.help
        if self.mode:
            string = string + " " + self.mode
        if self.file:
            string = string + " " + self.file
        elif self.querry:
            print(self.filename)
            with open(self.filename, "w") as f:
                f.write(self.querry)
            string = string + " " + str(self.filename)
        string = string + " --dialect " + self.dialect
        if self.rules:
            string = string + " --rules" + self.rules
        output = os.popen(string).read()
        if self.querry:
            os.remove(self.filename)
        return output
